/* expdbflib - library to export data from dbf files 
 * Copyright (C) 1997-1999 Zbigniew Michal Kempczynski
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* --------------------------------------- */
/*  Implementacja operacji na plikach DBF  */
/* --------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "dbf.h"

/* --------------------------------------------------------------------- */
/*	 TA CZESC PLIKU TO NISKOPOZIOMOWE ODWOLANIA DO PLIKOW DBF	 */
/* --------------------------------------------------------------------- */

/*			     HEADER FUNCTIONS				 */

/* Funkcja czyta naglowek pliku dbf - jesli sie powiedzie wynikiem jest 0
   w przeciwnym wypadku -1						 */
int DBF_fill_header(FILE *dbfdes, DBFheader *dbfh)
 {
  int succread;
  if (fseek(dbfdes, 0, SEEK_SET) != 0) return -1;
  if (fread(dbfh, sizeof(DBFheader), 1, dbfdes) > 0) return 0;
   else return -1;
 };

 /* Funkcja pokazuje informacje o naglowku dbf */
void DBF_show_header(DBFheader *dbfh)
{
 DBFdate t;
 printf("Identyfikator bazy	    : %d\n", DBF_get_id(dbfh));
 t = DBF_get_date(dbfh);
 printf("Data ostatniej modyfikacj  : "); DBF_show_date(&t); printf("\n");
 printf("Liczba rekordow	    : %d\n", DBF_get_recnum(dbfh));
 printf("Offset pierwszego rekordu  : %x\n", DBF_get_firstrecpos(dbfh));
 printf("Dlugosc rekordu	    : %x\n", DBF_get_reclen(dbfh));
}

 /* Pobranie identyfikatora bazy */
char DBF_get_id(DBFheader *dbfh) { return dbfh -> id; }

 /* Pobranie daty ostatniej modyfikacji bazy */
DBFdate DBF_get_date(DBFheader *dbfh)
{
  DBFdate tmp;
  tmp.year = dbfh -> date[0];
  (tmp.year < 80) ? tmp.year+=2000 : tmp.year+=1900;
  tmp.month = dbfh -> date[1];
  tmp.day = dbfh -> date[2];
  return tmp;
}

 /* Wyswietla date zawarta w formacie DBFdate */
void DBF_show_date(DBFdate *dbfdate)
{
 printf("%d:%d:%d", dbfdate -> year, dbfdate -> month, dbfdate -> day);
}

 /* Pobranie ilosci rekordow zawartych w bazie */
long DBF_get_recnum(DBFheader *dbfh) { return dbfh -> recnum; }

 /* Pobranie offsetu pierwszego rekordu w stosunku do poczatku pliku */
word DBF_get_firstrecpos(DBFheader *dbfh) { return dbfh -> firstrecpos; }

 /* Pobranie ilosci rekordow w bazie */
word DBF_get_reclen(DBFheader *dbfh) { return dbfh -> reclen; }

 /*			      BLOCK FUNCTIONS				 */
  /* Ustawienie wskaznika pliku na zadanym bloku: 0 = sukces, -1 blad	  */
int DBF_set_block(FILE *dbfdes, int blocknum)
{
 if (fseek(dbfdes, sizeof(DBFheader)+blocknum*sizeof(DBFblock), SEEK_SET) == 0) return 0;
  else return -1;
}

 /* Wczytanie bloku */
int DBF_read_block(FILE *dbfdes, DBFblock *dbfh)
{
 if (fread(dbfh, sizeof(DBFblock), 1, dbfdes) > 0) return 0;
  else return -1;
}

 /* Wyswietlenie informacji o bloku */
void DBF_show_block(DBFblock *dbfb)
{
 char txt[80];
 printf("Pole : %s ", dbfb -> fieldname);
// gotoxy(20, wherey());
 printf(" | ");
 if (dbfb -> fieldtype == 'C') printf("Char");
  else if (dbfb -> fieldtype == 'D') printf("Data");
   else if (dbfb -> fieldtype == 'L') printf("Logical");
    else if (dbfb -> fieldtype == 'M') printf("Memo");
     else if (dbfb -> fieldtype == 'N') printf("Numeric");
// gotoxy(30, wherey());
 printf(" | ");
 printf("Dlugosc : ");
 if (dbfb -> fieldtype == 'N') printf("%d:%d",
				      dbfb -> fieldlen[0],
				      dbfb -> fieldlen[1]);
  else printf("%d", dbfb -> fieldlen[0]);
 printf("\n");
}

 /* Sprawdzenie czy podany block jest blokiem koncowym: 0 - jest; -1 nie */
int DBF_is_end_of_blocks(DBFblock *dbfb)
{
 if (dbfb -> fieldname[0] == DBF_END_OF_BLOCKS) return 0;
  else return -1;
}

/* --------------------------------------------------------------------- */
/*	TA CZESC PLIKU TO SREDNIOPOZIOMOWE ODWOLANIA DO PLIKOW DBF	 */
/* --------------------------------------------------------------------- */

 /* Funkcja otwiera i wczytuje naglowek pliku dbf */
 /* Jezeli 0 to otwarcie pliku powiodlo sie	  */
int DBF_open_dbffile(DBFdescheader *dbfdh, char *dbfname)
{
  if ((dbfdh -> dbfdes = fopen(dbfname, "rb")) == NULL) fopen_error(dbfname);
  else
  {
   strcpy(dbfdh -> dbfname, dbfname);
   if (DBF_fill_header(dbfdh -> dbfdes, &dbfdh -> dbfh) != 0) return -1;
  }
 return 0;
}

 /* Wyswietlenie informacji o naglowku pliku dbf i nazwy pliku */
void DBF_show_dbffile(DBFdescheader *dbfdh)
{
 printf("Nazwa pliku		    : %s\n", dbfdh -> dbfname);
 DBF_show_header(&dbfdh -> dbfh);
}


 /* Jezeli 0 to zamkniecie pliku powiodlo sie */
int DBF_close_dbffile(DBFdescheader *dbfdh)
{
 if (fclose(dbfdh -> dbfdes) != 0) return -1;
  else strcpy(dbfdh -> dbfname, "");
 return 0;
}

 /* Funkcja zwraca liczbe blokow lub -1 w przypadku bledu */
int DBF_count_blocks(DBFdescheader *dbfdh)
{
 DBFblock blk;
 int cnt1;
 if (DBF_set_block(dbfdh -> dbfdes, 0) < 0) return -1;
 for (cnt1 = 0; cnt1 < 0x100; cnt1++)
  {
   if (DBF_read_block(dbfdh -> dbfdes, &blk) < 0) return -1;
    else if (DBF_is_end_of_blocks(&blk) == 0) break;
  }
 return cnt1;
}

 /* Funkcja tworzy tablice blokow pliku dbf */
void DBF_create_block_tree(DBFblocktree *dbfbt, DBFdescheader *dbfdh)
{
 int cnt1;
 int tmp;
 dbfbt -> blocknum = DBF_count_blocks(dbfdh);
 dbfbt -> dbfbtab = new DBFblock[dbfbt -> blocknum];
 dbfbt -> dbffieldpos = new short[dbfbt -> blocknum];
 DBF_set_block(dbfdh -> dbfdes, 0);
 for (cnt1 = 0; cnt1 < dbfbt -> blocknum; cnt1++)
  DBF_read_block(dbfdh -> dbfdes, &dbfbt -> dbfbtab[cnt1]);
 tmp = 0;
 for (cnt1 = 0; cnt1 < dbfbt -> blocknum; cnt1++)
  {
   dbfbt -> dbffieldpos[cnt1] = tmp;
    tmp += dbfbt -> dbfbtab[cnt1].fieldlen[0];
  }
}

 /* Funkcja usuwa tablice blokow */
void DBF_destroy_block_tree(DBFblocktree *dbfbt)
{
 delete dbfbt -> dbfbtab;
 delete dbfbt -> dbffieldpos;
}

void DBF_show_block_tree(DBFblocktree *dbfbt)
{
 int cnt1;
 printf("Liczba pol w bazie : %d\n", dbfbt -> blocknum);
 for (cnt1 = 0; cnt1 < dbfbt -> blocknum; cnt1++)
  DBF_show_block(&dbfbt -> dbfbtab[cnt1]);
}

void DBF_allocate_record(DBFdescheader *dbfdh, DBFrecord *dbfr)
{
 dbfr -> rec = (void *)new char[dbfdh -> dbfh.reclen + 1]; //przydzial pamieci
 ((char *) (dbfr -> rec))[dbfdh -> dbfh.reclen] = 0;
 DBF_set_record(dbfdh, dbfr, 0); //ustawienie wskaznika
}

void DBF_set_record(DBFdescheader *dbfdh, DBFrecord *dbfr, long rec)
{
 if (rec == 0)
     fseek(dbfdh -> dbfdes,
	   dbfdh -> dbfh.firstrecpos + 1,
	   SEEK_SET);

 else fseek(dbfdh -> dbfdes,
	  (rec - dbfr -> actrec) * dbfdh -> dbfh.reclen,
	  SEEK_CUR);
 dbfr -> actrec = rec;
}

void DBF_release_record(DBFrecord *dbfr)
{
 delete dbfr -> rec;
}

int DBF_read_record(DBFdescheader *dbfdh, DBFrecord *dbfr)
{
 if (fread(dbfr -> rec, dbfdh -> dbfh.reclen, 1, dbfdh -> dbfdes) == 0) return -1;
  else (dbfr -> actrec)++;
 return 0;
}

int DBF_get_field_len(DBFblocktree *dbfbt, int fieldnum)
{
 return dbfbt -> dbfbtab[fieldnum].fieldlen[0];
}

int DBF_get_field_num(DBFblocktree *dbfbt, char *fieldname)
{
 int cnt1, status;
 status = -1;
 for (cnt1 = 0; cnt1 < dbfbt -> blocknum; cnt1++)
   if (strcmp(dbfbt -> dbfbtab[cnt1].fieldname, fieldname) == 0)
      {
       status = 0;
       break;
      }
 if (status != 0) return -1;
 return cnt1;
}

void DBF_show_field(DBFblocktree *dbfbt, DBFrecord *rec, int fieldnum)
{
 int cnt1;
 for (cnt1 = 0; cnt1 < dbfbt -> dbfbtab[fieldnum].fieldlen[0]; cnt1++)
   printf("%c", ((char *)rec -> rec)[cnt1 + dbfbt -> dbffieldpos[fieldnum]]);
}

void DBF_show_field_by_fieldname(DBFblocktree *dbfbt, DBFrecord *rec, char *fieldname)
{
 int cnt1, status;
 status = -1;
 for (cnt1 = 0; cnt1 < dbfbt -> blocknum; cnt1++)
   if (strcmp(dbfbt -> dbfbtab[cnt1].fieldname, fieldname) == 0)
      {
       status = 0;
       break;
      }
 if (status == 0) DBF_show_field(dbfbt, rec, cnt1);
  else
    printf(" !!! %s not exist !!! ", fieldname);
}

DBFfield DBF_get_field_from_record(DBFblocktree *dbfbt, DBFrecord *rec, int fieldnum)
{
 DBFfield tmp;
 tmp.fieldlen = dbfbt -> dbfbtab[fieldnum].fieldlen[0];
 tmp.field = &((char *)rec -> rec)[dbfbt -> dbffieldpos[fieldnum]];
 return tmp;
}

DBFfield DBF_get_field_from_record_by_fieldname(DBFblocktree *dbfbt, DBFrecord *rec, char *fieldname)
{
 int cnt1, status;
 status = -1;
 for (cnt1 = 0; cnt1 < dbfbt -> blocknum; cnt1++)
   if (strcmp(dbfbt -> dbfbtab[cnt1].fieldname, fieldname) == 0)
      {
       status = 0;
       break;
      }
 if (status == 0) ;
  else
    printf(" !!! %s not exist !!! ", fieldname);
 return DBF_get_field_from_record(dbfbt, rec, cnt1);
}

/* ------------------------------------------------------------------- */
/* KLASA DBF implementacja					       */
/* ------------------------------------------------------------------- */
DBF::DBF(char *dbfname)
{
 ::DBF_open_dbffile(&dbfdh, dbfname);	   //otworzenie pliku
 ::DBF_create_block_tree(&dbfbt, &dbfdh);  //utworzenie tablicy blokow
 ::DBF_allocate_record(&dbfdh, &dbfr);	   //przydzielenie pamieci dla rekordu
 dbff.fieldlen = 0;
 dbff.field = NULL;
}

DBF::~DBF(void)
{
 ::DBF_release_record(&dbfr);		     //zwolnienie pamieci rekordu
 ::DBF_destroy_block_tree(&dbfbt);	     //zwolnienie pamieci dla pol
 ::DBF_close_dbffile(&dbfdh);
 delete dbff.field;
}
long DBF::DBF_get_recnum(void)	   //funkcja zwraca liczbe rekordow
{
 return ::DBF_get_recnum(&dbfdh.dbfh);
}

void DBF::DBF_read_record(long rec)
{
 ::DBF_set_record(&dbfdh, &dbfr, rec); //ustawienie wskaznika na zadany rekord
 ::DBF_read_record(&dbfdh, &dbfr);   //wczytanie rekordu
}


void DBF::DBF_get_field_by_num(int fieldnum)
{
 int tmplen;
 if (dbff.fieldlen > 0) delete dbff.field;
 tmplen = ::DBF_get_field_len(&dbfbt, fieldnum);
 dbff.field = new char[tmplen + 1];
 dbff.field[tmplen] = '\0';
 memcpy(dbff.field, &((char *)dbfr.rec)[dbfbt.dbffieldpos[fieldnum]], tmplen);
 dbff.fieldlen = strlen(dbff.field);
}

void DBF::DBF_get_field_by_name(char *fieldname)
{
 DBF_get_field_by_num(::DBF_get_field_num(&dbfbt, fieldname));
}

void DBF::DBF_show_field(void)
{
 printf("%s", (char *)dbff.field);
}

char *DBF::DBF_get_field_addr(void)
{
 return (char *)dbff.field;
}


void DBF::DBF_show_header(void)
{
 ::DBF_show_dbffile(&dbfdh);
}

void DBF::DBF_show_fields(void)
{
 ::DBF_show_block_tree(&dbfbt);
}

void DBF::DBF_show_field_by_num(int fieldnum)
{
 ::DBF_show_field(&dbfbt, &dbfr, fieldnum);
}

void DBF::DBF_show_field_by_name(char *fieldname)
{
 ::DBF_show_field_by_fieldname(&dbfbt, &dbfr, fieldname);
}

